#ifndef _KEYBOARD_HPP
#define _KEYBOARD_HPP

#include <keypad.hpp>
#include <Arduino.h>
#include "midi_interface.hpp"
#include "Key.hpp"
#include "chordprogression.hpp"

class Keyboard
{
public:
    KeyPad *chordKeys, *upperOctave, *lowerOctave;
    Midi *midi;
    uint8_t rootNote = 40;
    const uint8_t majorScaleIncrements[N_TONE_KEYS] = {0, 2, 4, 5, 7, 9, 11, 12, 14, 16, 17, 19, 21, 23, 24, 26};
    const uint8_t minorScaleIncrements[N_TONE_KEYS] = {0, 2, 3, 5, 7, 8, 10, 12, 14, 15, 17, 19, 20, 22, 24, 26};
    const uint8_t majorChordIntervals[4] = {4, 3, 4, 1};
    const uint8_t minorChordIntervals[4] = {3, 4, 3, 2};
    


    ChordProgression *chordProgression;
    uint8_t chordLeds[4] = {1, 0, 3, 2};

    Key *keys[16], *chords[4];

    uint8_t upperOctaveButtonOrder [8] = { 3, 4, 2, 5, 1, 6, 0, 7};
    uint8_t lowerOctaveButtonOrder[8] = {7, 0, 6, 1, 5, 2, 4, 3};
    uint8_t chordButtonOrder[8] = {3, 4, 5, 2, 6, 7, 0, 1};

    uint16_t sensorValues[4];
    uint16_t sensorMin[4] = {50, 170, 1, 200};
    uint16_t sensorMax[4] = {765, 538, 800,650};
    uint8_t lastPressureValue = 0;

    Keyboard(Midi *midi, ChordProgression *chordProgression)
    {
        upperOctave = new KeyPad(0x10, upperOctaveButtonOrder);
        lowerOctave = new KeyPad(0X20, lowerOctaveButtonOrder);
        chordKeys = new KeyPad(0x30, chordButtonOrder);

        this->midi = midi;

        for (uint8_t iKey = 0; iKey < N_TONE_KEYS; iKey++)
        {           
            KeyPad *keyPad = lowerOctave;
            if (iKey < 8)
                keyPad = upperOctave;

            Key *key = new Key(
                rootNote + majorScaleIncrements[iKey],
                keyPad, iKey % 8,
                midi,
                createColor(50, 50, 50),
                createColor(50, 50, 50));

            keys[iKey] = key;
        }
        
        for (uint8_t iKey = 0; iKey < N_CHORD_KEYS; iKey++)
        {            
            Key *key = new Key(
                rootNote + chordProgression->distances[iKey],
                chordKeys, chordLeds[iKey],
                midi,
                createColor(50, 50, 0),
                createColor(50, 50, 50));

            key->isChord = true;
            key->chordQuality = chordProgression->qualities[iKey];

            chords[iKey] = key;
        }
        //this->chordProgression = chordProgression;
        setChordProgression(chordProgression);
    }

    void setChordProgression(ChordProgression *chordProgression){
        this->chordProgression = chordProgression;
        remapChords(chordProgression);
        remapKeys(chordProgression->qualities[0], rootNote);
    }

    void remapChords(ChordProgression *chordProgression){
        for (uint8_t iKey = 0; iKey < N_CHORD_KEYS; iKey++){
            chords[iKey]->chordQuality = chordProgression->qualities[iKey];
            chords[iKey]->note = rootNote + chordProgression->distances[iKey];
            if (chords[iKey]->chordQuality == 'm') chords[iKey]->activeColor = createColor(0,50,0);
            if (chords[iKey]->chordQuality == 'M') chords[iKey]->activeColor = createColor(0,0,50);  
            chords[iKey]->activate();          
        }

        remapKeys(chordProgression->qualities[0], rootNote);
    }

    void remapKeys(char quality, uint8_t note)
    {
        const uint8_t *scale;
        uint8_t r,g, b;
        const uint8_t *chords;

        if (quality == 'm'){
            scale = minorScaleIncrements;
            chords = minorChordIntervals;            
        } else{
            scale = majorScaleIncrements;
            chords = majorChordIntervals;            
        }

        uint8_t keyNote = note;
        for (uint8_t iKey = 0; iKey < N_TONE_KEYS; iKey++)
        {
            uint8_t shade = map(keyNote, 31, 127, 0, 127);
            keys[iKey]->note = keyNote;
            if (quality == 'm') keys[iKey]->activeColor = createColor(0, 50, shade);
            if (quality == 'M') keys[iKey]->activeColor = createColor(shade, 0, 50);
            keys[iKey]->activate();
            keyNote += chords[iKey%4];
        }            
    }

    void update()
    {
        chordKeys->update();
        upperOctave->update();
        lowerOctave->update();

        for (int iKey = 0; iKey < N_KEYPAD_KEYS; iKey++)
        {
            if (upperOctave->getButtonPressed(iKey)) keys[iKey]->noteOn(127);               
            if (upperOctave->getButtonReleased(iKey)) keys[iKey]->noteOff();

            if (lowerOctave->getButtonPressed(iKey)) keys[iKey + 8]->noteOn(127);
            if (lowerOctave->getButtonReleased(iKey)) keys[iKey + 8]->noteOff();            
        }

        for (int iKey = 0; iKey < N_CHORD_KEYS; iKey++)
        {
            
            if (chordKeys->getButtonPressed(iKey))
            {
                midi->midiPanic();
                Key *chordKey = chords[iKey];

                chords[iKey]->noteOn(10);
                remapKeys(chordKey->chordQuality, chordKey->note-12);
                

            }

            if (chordKeys->getButtonReleased(iKey))
                chords[iKey]->noteOff();
        }

        sensorValues[0] = map(chordKeys->inputState.sensor1, sensorMin[0], sensorMax[0], 10, 100);
        sensorValues[1] = map(chordKeys->inputState.sensor2, sensorMin[1], sensorMax[1], 10, 100);
        sensorValues[2] = map(upperOctave->inputState.sensor1, sensorMin[2], sensorMax[2], 10, 100);
        sensorValues[3] = map(lowerOctave->inputState.sensor2, sensorMin[3], sensorMax[3], 10, 100);

        uint8_t pressure = constrain(max(sensorValues[0], max(sensorValues[1], max(sensorValues[2], sensorValues[3]))), 0, 127);
        if (abs(pressure - lastPressureValue) > 20){
            //Serial.println(pressure);
           // midi->sendPressure(pressure, 1);
        }
        lastPressureValue = pressure;
    }

    String toString()
    {
        return String("\nChordkeys") + chordKeys->toString() +
               String("\nupperOctave") + upperOctave->toString() +
               String("\nlowerOctave") + lowerOctave->toString();
    }

    void clearLEDs()
    {
        chordKeys->clearLEDs();
        upperOctave->clearLEDs();
        lowerOctave->clearLEDs();
    }
};

#endif